# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Import classes, functions, and submodules to be accessible from library
"""

from accustom.constants import Status
from accustom.constants import RequestType
from accustom.constants import RedactMode
from accustom.response import ResponseObject
from accustom.response import cfnresponse
from accustom.response import is_valid_event
from accustom.response import collapse_data
from accustom.decorators import decorator
from accustom.decorators import rdecorator
from accustom.decorators import sdecorator
from accustom.redaction import RedactionConfig
from accustom.redaction import StandaloneRedactionConfig
from accustom.redaction import RedactionRuleSet

import accustom.Exceptions



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Exceptions for the accustom library.

These are the exceptions that can be returned by accustom
"""


class CannotApplyRuleToStandaloneRedactionConfig(Exception):
    """Indicates that a second rule set was attempted to be applied to a standalone"""

    def __init__(self, *args):
        Exception.__init__(self, *args)


class ConflictingValue(Exception):
    """Indicates that there is already a record with this value"""

    def __init__(self, *args):
        Exception.__init__(self, *args)


class NoPhysicalResourceIdException(Exception):
    """Indicates that there was no valid value to use for PhysicalResourceId"""

    def __init__(self, *args):
        Exception.__init__(self, *args)


class InvalidResponseStatusException(Exception):
    """Indicates that there response code was not SUCCESS or FAILED"""

    def __init__(self, *args):
        Exception.__init__(self, *args)


class DataIsNotDictException(Exception):
    """Indicates that a Dictionary was not passed as Data"""

    def __init__(self, *args):
        Exception.__init__(self, *args)


class FailedToSendResponseException(Exception):
    """Indicates there was a problem sending the response"""

    def __init__(self, *args):
        Exception.__init__(self, *args)


class NotValidRequestObjectException(Exception):
    """Indicates that the event passed in is not a valid Request Object"""

    def __init__(self, *args):
        Exception.__init__(self, *args)


class ResponseTooLongException(Exception):
    """Indicates that the produced response exceeds 4096 bytes and thus is too long"""

    def __init__(self, *args):
        Exception.__init__(self, *args)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

"""
Bring exceptions into the root of the submodule
"""

from accustom.Exceptions.exceptions import CannotApplyRuleToStandaloneRedactionConfig
from accustom.Exceptions.exceptions import ConflictingValue
from accustom.Exceptions.exceptions import NoPhysicalResourceIdException
from accustom.Exceptions.exceptions import InvalidResponseStatusException
from accustom.Exceptions.exceptions import DataIsNotDictException
from accustom.Exceptions.exceptions import FailedToSendResponseException
from accustom.Exceptions.exceptions import NotValidRequestObjectException
from accustom.Exceptions.exceptions import ResponseTooLongException



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

""" RedactionConfig and StandaloneRedactConfig for the accustom library

This allows you to define a redaction policy for accustom
"""

from accustom.response import is_valid_event
from accustom.constants import RedactMode
from accustom.Exceptions import ConflictingValue
from accustom.Exceptions import CannotApplyRuleToStandaloneRedactionConfig
from accustom.Exceptions import NotValidRequestObjectException

import logging
import six
import re
import copy

logger = logging.getLogger(__name__)

_RESOURCEREGEX_DEFAULT = '^.*$'
REDACTED_STRING = '[REDACTED]'


class RedactionRuleSet(object):
    """Class that allows you to define a redaction rule set for accustom"""

    def __init__(self, resourceRegex: str = _RESOURCEREGEX_DEFAULT):
        """Init function for the class

        Args:
            resourceRegex (String): The regex used to work out what resources to apply this too.

        Raises:
            TypeError

        """
        if not isinstance(resourceRegex, six.string_types):
            raise TypeError('resourceRegex must be a string')

        self.resourceRegex = resourceRegex
        self._properties = []

    def add_property_regex(self, propertiesRegex: str):
        """Allows you to add a property regex to allowlist/blocklist

        Args:
            propertiesRegex (String): The regex used to work out what properties to allowlist/blocklist

        Raises:
            TypeError

        """
        if not isinstance(propertiesRegex, six.string_types):
            raise TypeError('propertiesRegex must be a string')
        self._properties.append(propertiesRegex)

    def add_property(self, propertyName: str):
        """Allows you to add a specific property to allowlist/blocklist

        Args:
            propertyName (String): The name of the property to allowlist/blocklist

        Raises:
            TypeError

        """
        if not isinstance(propertyName, six.string_types):
            raise TypeError('propertyName must be a string')
        self._properties.append('^' + propertyName + '$')


class RedactionConfig(object):
    """Class that defines a redaction policy for accustom"""

    def __init__(self, redactMode: str = RedactMode.BLOCKLIST, redactResponseURL: bool = False):
        """Init function for the class

        Args:
            redactMode (RedactMode.BLOCKLIST or RedactMode.ALLOWLIST): Determine if we should allowlist or blocklist
            resources, defaults to blocklist
        redactResponseURL (boolean): Prevents the pre-signed URL from being printed preventing out of band responses
        (recommended for production)

        Raises:
            TypeError

        """
        if redactMode == RedactMode.BLACKLIST:
            logger.warning("The usage of RedactMode.BLACKLIST is deprecated, "
                           "please change to use RedactMode.BLOCKLIST")
            redactMode = RedactMode.BLOCKLIST
        if redactMode == RedactMode.WHITELIST:
            logging.warning("The usage of RedactMode.WHITELIST is deprecated, "
                            "please change to use RedactMode.ALLOWLIST")
            redactMode = RedactMode.ALLOWLIST

        if redactMode != RedactMode.BLOCKLIST and redactMode != RedactMode.ALLOWLIST:
            raise TypeError('Invalid Redaction Type')

        if not isinstance(redactResponseURL, bool):
            raise TypeError('redactResponseURL must be of boolean type')

        self.redactMode = redactMode
        self.redactResponseURL = redactResponseURL
        self._redactProperties = {}

    def add_rule_set(self, ruleSet: RedactionRuleSet):
        """ This function will add a RedactionRuleSet object to the RedactionConfig.

        Args:
            ruleSet (RedactionRuleSet): The rule to be added to the RedactionConfig

        Raises:
            TypeError
            ConflictingValue

        """

        if not isinstance(ruleSet, RedactionRuleSet):
            raise TypeError('Please use RedactionRuleSet class')
        if ruleSet.resourceRegex in self._redactProperties:
            raise ConflictingValue(f'There is already a record set for resource of regex: {ruleSet.resourceRegex}')

        # noinspection PyProtectedMember
        self._redactProperties[ruleSet.resourceRegex] = ruleSet._properties

    def _redact(self, event: dict):
        """ Internal Function. Not to be consumed outside accustom Library.

            This function will take in an event and return the event redacted as per the redaction config.
        """
        if not is_valid_event(event):
            # If it is not a valid event we need to raise an exception
            message = 'The event object passed is not a valid Request Object as per ' + \
                      'https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/crpg-ref-requests.html'
            logger.error(message)
            raise NotValidRequestObjectException(message)
        ec = copy.deepcopy(event)
        if self.redactMode == RedactMode.ALLOWLIST:
            if 'ResourceProperties' in ec: ec['ResourceProperties'] = {}
            if 'OldResourceProperties' in ec: ec['OldResourceProperties'] = {}
        for resourceRegex, propertyRegex in self._redactProperties.items():
            if re.search(resourceRegex, event['ResourceType']) is not None:
                # Go through the Properties looking to see if they're in the ResourceProperties or OldResourceProperties
                for index, item in enumerate(propertyRegex):
                    r = re.compile(item)
                    if self.redactMode == RedactMode.BLOCKLIST:
                        if 'ResourceProperties' in ec:
                            for m_item in filter(r.match, ec['ResourceProperties']):
                                ec['ResourceProperties'][m_item] = REDACTED_STRING
                        if 'OldResourceProperties' in ec:
                            for m_item in filter(r.match, ec['OldResourceProperties']):
                                ec['OldResourceProperties'][m_item] = REDACTED_STRING
                    elif self.redactMode == RedactMode.ALLOWLIST:
                        if 'ResourceProperties' in ec:
                            for m_item in filter(r.match, event['ResourceProperties']):
                                ec['ResourceProperties'][m_item] = event['ResourceProperties'][m_item]
                        if 'OldResourceProperties' in ec:
                            for m_item in filter(r.match, event['OldResourceProperties']):
                                ec['OldResourceProperties'][m_item] = event['OldResourceProperties'][m_item]
        if self.redactMode == RedactMode.ALLOWLIST:
            if 'ResourceProperties' in ec:
                for key, value in event['ResourceProperties'].items():
                    if key not in ec['ResourceProperties']: ec['ResourceProperties'][key] = REDACTED_STRING
            if 'OldResourceProperties' in ec:
                for key, value in event['OldResourceProperties'].items():
                    if key not in ec['OldResourceProperties']: ec['OldResourceProperties'][key] = REDACTED_STRING

        if self.redactResponseURL: del ec['ResponseURL']
        return ec

    def __str__(self):
        return f'RedactionConfig({self.redactMode})'

    def __repr__(self):
        return str(self)


class StandaloneRedactionConfig(RedactionConfig):
    """
    Class that defines a redaction policy for a standalone function
    """
    def __init__(self, ruleSet: RedactionRuleSet, redactMode: str = RedactMode.BLOCKLIST,
                 redactResponseURL: bool = False):
        """Init function for the class

        Args:
            redactMode (RedactMode.BLOCKLIST or RedactMode.ALLOWLIST): Determine if we should allowlist or blocklist
            resources, defaults to blocklist
        redactResponseURL (boolean): Prevents the pre-signed URL from being printed preventing out of band responses
        (recommended for production)
        ruleSet (RedactionRuleSet): The single rule set to be applied

        Raises:
            TypeError
        """

        RedactionConfig.__init__(self, redactMode=redactMode, redactResponseURL=redactResponseURL)
        ruleSet.resourceRegex = _RESOURCEREGEX_DEFAULT
        # override resource regex to be default
        assert (ruleSet is not None)
        RedactionConfig.add_rule_set(self, ruleSet)

    def add_rule_set(self, ruleSet: RedactionRuleSet):
        """ Overrides the add_rule_set operation with one that will immediately throw an exception

            Raises
                CannotApplyRuleToStandaloneRedactionConfig
        """
        raise CannotApplyRuleToStandaloneRedactionConfig('StandaloneRedactionConfig does not support additional rules.')



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0

from setuptools import setup
import os


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name='accustom',
    version='1.2.1',
    description='Accustom is a library for responding to Custom Resources in AWS CloudFormation using the decorator '
                'pattern.',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/awslabs/cloudformation-accustom',
    author='Taylor Bertie',
    author_email='bertiet@amazon.com',
    license='Apache-2.0',
    download_url='https://github.com/awslabs/cloudformation-accustom/archive/1.1.2.tar.gz',
    keywords=['cloudformation', 'lambda', 'custom', 'resource', 'decorator'],
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Application Frameworks',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],
    packages=['accustom', 'accustom.Exceptions'],
    install_requires=[
        'botocore>=1.29',
        'boto3>=1.26',
        'requests>=2.25',
        'six>=1.16'
    ],
    python_requires='>=3, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, !=3.6.*, <4',
)
